/*
    Saturn SCU DSP disassembler
    by Anders Montonen
    WWW: http://www.helsinki.fi/~ammonton/sega/
    
    Copyright (C) 2003 Anders Montonen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include "ops.h"

typedef unsigned int u32;

u32 readword(void);
void aluops(u32 word);
void mvi(u32 word);
void ctrl(u32 word);

FILE *infile = NULL;
FILE *outfile = NULL;

int filestatus = 0;

int main(int argc, char *argv[])
{
    int row = 0;
    u32 word;
    
    if (argc < 2)
    {
        printf("Usage: scu-dump input [output]\n");
        return 0;
    }
    infile = fopen(argv[1], "rb");
    if (infile == NULL)
    {
        printf("Couldn't open file for input.\n");
        return 0;
    }
    if (argc == 2)
    {
        outfile = stdout;
    }
    else
    {
        outfile = fopen(argv[2], "w");
        if (outfile == NULL)
        {
            printf("Couldn't open file for output.\n");
            fclose(infile);
            return 0;
        }
    }
    while (!filestatus)
    {
        word = readword();
        if (!filestatus)
        {
            fprintf(outfile, "%03x: %08lx   ", row, word);
            switch (word >> 30) {
            case 0:
                aluops(word);
                break;
            case 2:
                mvi(word);
                break;
            case 3:
                ctrl(word);
                break;
            default:
                fprintf(outfile, "unrecognized or data\n");
            }
            
        }
        else
            break;
        row++;
    }
    fclose(infile);
    fclose(outfile);
    return 0;
}

void aluops(u32 word)
{
	const char *apad = "     ";
	const char *pad1 = "           ";
    int len;
    /* A-bus */
	len = 0;
    len += fprintf(outfile, "%s", abus_ops[(word >> 26) & 0xf]);
    fprintf(outfile, "%s", apad+len);
    /* X-bus */
    if (!(word & 0x3800000))
    {
		len = 0;
        len += fprintf(outfile, "nop");
		fprintf(outfile,"%s%s", pad1+len, pad1);
    }
    else
    {
        if ((word >> 23 & 0x7) == 3)
        {
			len = 0;
			fprintf(outfile, "%s", pad1);
            len += fprintf(outfile, "mov %s,P", xy_args[(word >> 20) & 0x7]);
			fprintf(outfile, "%s", pad1+len); 
        }
        else
        {
			len = 0;
            if (word & 0x2000000)
				len += fprintf(outfile, "mov %s,X", xy_args[(word >> 20) & 0x7]);
            fprintf(outfile, "%s", pad1+len);
			len = 0;
            if ((word >> 23 & 3) == 2)
                len += fprintf(outfile, "mov MUL,P");
			fprintf(outfile, "%s", pad1+len);
        }
    }
    /* Y-bus */
    len = 0;
    if (!(word & 0xE0000))
    {
		len = 0;
        len += fprintf(outfile, "nop");
		fprintf(outfile, "%s%s", pad1+len, pad1);
    }
    else
    {
        if ((word >> 17 & 0x7) == 3)
        {
			len = 0;
			fprintf(outfile, "%s", pad1);
            len += fprintf(outfile, "mov %s,A", xy_args[(word >> 14) & 0x7]);
			fprintf(outfile, "%s", pad1+len);
        }
        else
        {
			len = 0;
            if (word & 0x80000)
                len += fprintf(outfile, "mov %s,Y", xy_args[(word >> 14) & 0x7]);
			fprintf(outfile, "%s", pad1+len);
			len = 0;
            if ((word >> 17 & 0x3) == 1)
                len += fprintf(outfile, "clr A");
            if ((word >> 17 & 0x3) == 2)
                len += fprintf(outfile, "mov ALU,A");
			fprintf(outfile, "%s", pad1+len);
        }
    }
    /* D1-bus */
    if ((word >> 12 & 0x3) == 0)
        fprintf(outfile, "nop");
    if ((word >> 12 & 0x3) == 1)
        fprintf(outfile, "mov %x,%s", (word & 0xff), d1movdest_args[(word >> 8) & 0xf]);
    if ((word >> 12 & 0x3) == 3)
        fprintf(outfile, "mov %s,%s", d1movsrc_args[word & 0xf], d1movdest_args[(word >> 8) & 0xf]);
    fprintf(outfile, "\n");
}

void mvi(u32 word)
{
    fprintf(outfile, "mvi ");
    if (!(word & 0x2000000))
    {
        fprintf(outfile, "%x,%s", (word & 0x1FFFFFF), mvi_args[(word >> 26) & 0xf]);
    }
    else
    {
        fprintf(outfile, "%x,%s,", (word & 0x7ffff), mvi_args[(word >> 26) & 0xf]);
        if (!(word & 0x1000000))
            fprintf(outfile, "N");
        switch ((word >> 19) & 0x1f) {
        case 0x1 :
            fprintf(outfile, "Z");
            break;
        case 0x2 :
            fprintf(outfile, "S");
            break;
        case 0x3 :
            fprintf(outfile, "ZS");
            break;
        case 0x4 :
            fprintf(outfile, "C");
            break;
        case 0x8 :
            fprintf(outfile, "T0");
            break;
        default:
            fprintf(outfile, "???");
        }
    }
    fprintf(outfile, "\n");
}

void ctrl(u32 word)
{
    if ((word >> 27) == 0x1c)
        fprintf(outfile, "btm");
    else if ((word >> 27) == 0x1d)
        fprintf(outfile, "lps");
    else if ((word >> 27) == 0x1e)
        fprintf(outfile, "end");
    else if ((word >> 27) == 0x1f)
        fprintf(outfile, "endi");
    else if ((word >> 28) == 0xc)
    {
        fprintf(outfile, "dma");
        if (word & 0x4000)
            fprintf(outfile, "h");
        fprintf(outfile, "%s ", dma_addr_add[(word >> 15) & 0x7]);
        if (word & 0x1000)
        {
            fprintf(outfile, "%s,D0,", dma1_args[(word >> 8) & 0x7]);
        }
        else
        {
            fprintf(outfile, "D0,%s,", dma1_args[(word >> 8) & 0x7]);
        }
        if (word & 0x2000)
        {
            fprintf(outfile, "%s", dma2_args[word & 0x7]);
        }
        else
        {
            fprintf(outfile, "%x", word & 0xff);
        }
    }
    else if ((word >> 28) == 0xd)
    {
        fprintf(outfile, "jmp ");
        if (word & 0x2000000)
        {
            if (!(word & 0x1000000))
                fprintf(outfile, "N");
            switch ((word >> 19) & 0x1f) {
            case 0x1 :
                fprintf(outfile, "Z,");
                break;
            case 0x2 :
                fprintf(outfile, "S,");
                break;
            case 0x3 :
                fprintf(outfile, "ZS,");
                break;
            case 0x4 :
                fprintf(outfile, "C,");
                break;
            case 0x8 :
                fprintf(outfile, "T0,");
                break;
            default:
                fprintf(outfile, "???,");
            }
        }
        fprintf(outfile, "%x", word & 0xff);
    }
    else
        fprintf(outfile, "???");
    fprintf(outfile, "\n");
}

/* Read a 32-bit big-endian word */
u32 readword(void)
{
    u32 word = 0;
    word = ((fgetc(infile) & 0xff) << 24);
    word |= ((fgetc(infile) & 0xff) << 16);
    word |= ((fgetc(infile) & 0xff) << 8);
    word |= (fgetc(infile) & 0xff);
    if (feof(infile) || ferror(infile))
        filestatus = 1;
    return word;
}
